
#include "util/slam_util.h"

namespace slam{

Quaterniond SlamUtil::EulerToQuaternion(double roll, double pitch, double yaw) {
    Quaterniond q;

    // 计算半角和它们的正余弦值
    double cy = cos(yaw * 0.5);
    double sy = sin(yaw * 0.5);
    double cp = cos(pitch * 0.5);
    double sp = sin(pitch * 0.5);
    double cr = cos(roll * 0.5);
    double sr = sin(roll * 0.5);

    // 计算四元数
    q.w = cr * cp * cy + sr * sp * sy;
    q.x = sr * cp * cy - cr * sp * sy;
    q.y = cr * sp * cy + sr * cp * sy;
    q.z = cr * cp * sy - sr * sp * cy;

    return q;
}

SlamUtil::slam_kd_tree_t* SlamUtil::buildKDTree(const std::vector<slam::Vector2d>& ref_line)
{
    auto* cloud = new PointCloud<double>(ref_line);
    auto* index = new slam_kd_tree_t(2 /*dim*/, *cloud, KDTreeSingleIndexAdaptorParams(10 /* max leaf */));
    index->buildIndex();
    return index;
}

// 查找最近点的函数
int SlamUtil::findNearest(slam_kd_tree_t* kd_tree, const slam::Vector2d& query_point, double& out_min_dist)
{
    const size_t num_results = 1;
    size_t ret_index;
    double out_dist_sqr;
    nanoflann::KNNResultSet<double> resultSet(num_results);
    resultSet.init(&ret_index, &out_dist_sqr);  // 初始化结果集
    double query_pt[2] = { query_point.x, query_point.y };  // 查询点

    // 构造搜索参数
    SearchParameters searchParams(10, true); // 这里我修改为 SearchParams

    // 执行搜索
    kd_tree->findNeighbors(resultSet, query_pt, searchParams);  // 使用SearchParams调用

    out_min_dist = sqrt(out_dist_sqr);  // 计算并返回实际距离（开方平方距离）
    return static_cast<int>(ret_index);  // 返回最近邻点的索引
}


// 计算两个 Vector2d 之间距离的平方
double squaredDistanceTo(const slam::Vector2d& a, const slam::Vector2d& b) {
    return (a.x - b.x) * (a.x - b.x) + (a.y - b.y) * (a.y - b.y);
}

slam::Vector2d SlamUtil::globalToLocal(slam::Vector2d glb_p,slam::Vector2d center_p, double heading) {

    slam::Vector2d res_p;

    double dev_x = glb_p.x - center_p.x;
    double dev_y = glb_p.y - center_p.y;

    res_p.x = dev_x * cos(heading) - dev_y*sin(heading);
    res_p.y = dev_x * sin(heading) + dev_y*cos(heading);

    return res_p;
}

slam::Vector2d SlamUtil::gcs2vcs(slam::Vector2d glb_p,slam::Vector2d center_p, double heading)
{
    slam::Vector2d res_p;

    double dev_x = glb_p.x - center_p.x;
    double dev_y = glb_p.y - center_p.y;

    res_p.x = dev_x * cos(heading) - dev_y*sin(heading);
    res_p.y = dev_x * sin(heading) + dev_y*cos(heading);

    return res_p;
}


// 查找最近点的函数
int SlamUtil::findNearestPointIndex(const slam::Vector2d& cur_host_pos, const std::vector<slam::Vector2d>& ref_line) {
    if (ref_line.empty()) {
        // 如果ref_line是空的，返回-1表示没有找到
        return -1;
    }

    int nearestIndex = 0; // 假设第一个点最近
    double minDistance = squaredDistanceTo(ref_line[0], cur_host_pos); // 初始化最小距离为第一个点的距离

    for (size_t i = 1; i < ref_line.size(); ++i) {
        double distance = squaredDistanceTo(ref_line[i], cur_host_pos);
        if (distance < minDistance) {
            // 如果找到了更近的点，更新最小距离和索引
            minDistance = distance;
            nearestIndex = i;
        }
    }

    return nearestIndex;
}

int SlamUtil::findNearestPointIndex(const slam::Vector2d& cur_host_pos,
                                    const std::vector<slam::Vector2d>& ref_line,
                                    double max_distance, double min_distance) {
    if (ref_line.empty()) return -1;

    // 使用平方距离来避免在比较时计算平方根
    double max_distance_sq = max_distance * max_distance;
    double min_distance_sq = min_distance * min_distance;

    int left = 0;
    int right = ref_line.size() - 1;
    int nearestIndex = -1;
    double nearestDistanceSq = std::numeric_limits<double>::max();

    while (left <= right) {
        int mid = left + (right - left) / 2;
        double midDistanceSq = squaredDistanceTo(ref_line[mid], cur_host_pos);

        // 更新最近点和距离
        if (midDistanceSq < nearestDistanceSq) {
            nearestDistanceSq = midDistanceSq;
            nearestIndex = mid;
        }

        // 检查是否已找到足够近的点
        if (nearestDistanceSq <= min_distance_sq) {
            return nearestIndex; // 找到足够近的点
        }

        // 比较首尾和中间点，决定下一步搜索方向
        double leftDistanceSq = squaredDistanceTo(ref_line[left], cur_host_pos);
        double rightDistanceSq = squaredDistanceTo(ref_line[right], cur_host_pos);

        if (leftDistanceSq < midDistanceSq) {
            right = mid - 1;
        } else if (rightDistanceSq < midDistanceSq) {
            left = mid + 1;
        } else {
            // 当左、中、右距离相似时，缩小搜索范围
            left++;
            right--;
        }
    }

    return nearestDistanceSq > max_distance_sq ? -1 : nearestIndex;
}


}
